from .maze import Square, random_maze, plot_maze, search_, search_animated, search_solution


# Se vuoi esporlo con il wildcard (*)
__all__ = ["Square","random_maze", "plot_maze", "search_", "search_animated", "searh_solution" ]